import copy
import datetime
import decimal
from typing import MutableSequence, Optional, Type, TypeVar
from .generated import cost_spec
from .cost_component import CostComponent
from . import internal
from .date import Date
from .punctuation import Asterisk
from .escaped_string import EscapedString
from .currency import Currency
from .number_expr import NumberExpr
from .amount import Amount
from .compound_amount import CompoundAmount
from .cost import UnitCost, TotalCost

_Self = TypeVar('_Self', bound='CostSpec')


@internal.tree_model
class CostSpec(cost_spec.CostSpec):
    
    @internal.custom_property
    def raw_cost_components(self) -> MutableSequence[CostComponent]:
        return self.raw_cost.raw_components

    raw_compound_amount_comp = internal.unordered_node_property(raw_cost_components, CompoundAmount, prepend=True)
    raw_amount_comp = internal.unordered_node_property(raw_cost_components, Amount, prepend=True)
    raw_number_comp = internal.unordered_node_property(raw_cost_components, NumberExpr, prepend=True)
    raw_currency_comp = internal.unordered_node_property(raw_cost_components, Currency, prepend=True)
    raw_date_comp = internal.unordered_node_property(raw_cost_components, Date)
    raw_label_comp = internal.unordered_node_property(raw_cost_components, EscapedString)
    raw_asterisk_comp = internal.unordered_node_property(raw_cost_components, Asterisk)

    @internal.custom_property
    def raw_number_per(self) -> Optional[NumberExpr]:
        if compound_amount := self.raw_compound_amount_comp:
            return compound_amount.raw_number_per
        if not isinstance(self.raw_cost, UnitCost):
            return None
        if amount := self.raw_amount_comp:
            return amount.raw_number
        return self.raw_number_comp

    @raw_number_per.setter
    def __raw_number_per(self, value: Optional[NumberExpr]) -> None:
        if compound_amount := self.raw_compound_amount_comp:  # CompoundAmount
            compound_amount.raw_number_per = value
        elif isinstance(self.raw_cost, UnitCost):
            if amount := self.raw_amount_comp:
                if value:  # Amount(per)
                    amount.raw_number = value
                else:  # Amount(per) - Number(per) -> Currency
                    self.raw_currency_comp = copy.deepcopy(amount.raw_currency)
                    self.raw_amount_comp = None
            elif (currency := self.raw_currency_comp) and value:  # Currency + Number(per) -> Amount(per)
                self.raw_amount_comp = Amount.from_children(value, copy.deepcopy(currency))
                self.raw_currency_comp = None
            else:  # Number(per)
                self.raw_number_comp = value
        elif isinstance(self.raw_cost, TotalCost) and value:
            if amount := self.raw_amount_comp:  # Amount(total) + Number(per) -> CompoundAmount
                self._into_unit_cost(self.raw_cost)
                compound_amount = CompoundAmount.from_children(
                    value,
                    copy.deepcopy(amount.raw_number),
                    copy.deepcopy(amount.raw_currency))
                self.raw_compound_amount_comp = compound_amount
                self.raw_amount_comp = None
            elif currency := self.raw_currency_comp:  # Currency(total) + Number(per) -> Amount(per)
                self._into_unit_cost(self.raw_cost)
                amount = Amount.from_children(value, copy.deepcopy(currency))
                self.raw_amount_comp = amount
                self.raw_currency_comp = None
            elif self.raw_number_comp:  # Number(total) + Number(per) -> error
                raise ValueError('Cannot set both number_per and number_total without a currency.')
            else:  # /(total) + Number(per) -> Number(per)
                self._into_unit_cost(self.raw_cost)
                self.raw_number_comp = value

    @internal.custom_property
    def raw_number_total(self) -> Optional[NumberExpr]:
        if compound_amount := self.raw_compound_amount_comp:
            return compound_amount.raw_number_total
        if not isinstance(self.raw_cost, TotalCost):
            return None
        if amount := self.raw_amount_comp:
            return amount.raw_number
        return self.raw_number_comp

    @raw_number_total.setter
    def __raw_number_total(self, value: Optional[NumberExpr]) -> None:
        if compound_amount := self.raw_compound_amount_comp:  # CompoundAmount
            compound_amount.raw_number_total = value
        elif isinstance(self.raw_cost, TotalCost):
            if amount := self.raw_amount_comp:
                if value:  # Amount(total):
                    amount.raw_number = value
                else:  # Amount(total) - Number(total) -> Currency
                    self.raw_currency_comp = copy.deepcopy(amount.raw_currency)
                    self.raw_amount_comp = None
            elif (currency := self.raw_currency_comp) and value:  # Currency + Number(total) -> Amount(total)
                self.raw_amount_comp = Amount.from_children(value, copy.deepcopy(currency))
                self.raw_currency_comp = None
            else:  # Number(total)
                self.raw_number_comp = value
        elif isinstance(self.raw_cost, UnitCost) and value:
            if amount := self.raw_amount_comp:  # Amount(per) + Number(total) -> CompoundAmount
                compound_amount = CompoundAmount.from_children(
                    copy.deepcopy(amount.raw_number),
                    value,
                    copy.deepcopy(amount.raw_currency))
                self.raw_compound_amount_comp = compound_amount
                self.raw_amount_comp = None
            elif currency := self.raw_currency_comp:  # Currency(per) + Number(total) -> Amount(total)
                self._into_total_cost(self.raw_cost)
                amount = Amount.from_children(value, copy.deepcopy(currency))
                self.raw_amount_comp = amount
                self.raw_currency_comp = None
            elif self.raw_number_comp:  # Number(per) + Number(total) -> error
                raise ValueError('Cannot set both number_per and number_total without a currency.')
            else:  # /(per) + Number(total) -> Number(total)
                self._into_total_cost(self.raw_cost)
                self.raw_number_comp = value

    @internal.custom_property
    def raw_currency(self) -> Optional[Currency]:
        if compound_amount := self.raw_compound_amount_comp:
            return compound_amount.raw_currency
        if amount := self.raw_amount_comp:
            return amount.raw_currency
        return self.raw_currency_comp

    @raw_currency.setter
    def __raw_currency(self, value: Optional[Currency]) -> None:
        if compound_amount := self.raw_compound_amount_comp:
            if value:  # CompoundAmount
                compound_amount.raw_currency = value
            else:  # CompoundAmount - Currency
                match compound_amount.raw_number_per, compound_amount.raw_number_total, self.raw_cost:
                    case NumberExpr(), None, UnitCost():
                        self.raw_number_comp = copy.deepcopy(compound_amount.raw_number_per)
                    case NumberExpr(), None, TotalCost() as c:
                        self._into_unit_cost(c)
                        self.raw_number_comp = copy.deepcopy(compound_amount.raw_number_per)
                    case None, NumberExpr(), TotalCost():
                        self.raw_number_comp = copy.deepcopy(compound_amount.raw_number_total)
                    case None, NumberExpr(), UnitCost() as c:
                        self._into_total_cost(c)
                        self.raw_number_comp = copy.deepcopy(compound_amount.raw_number_total)
                    case NumberExpr(), NumberExpr(), _:
                        raise ValueError('Cannot remove currency from compound amount with both numbers.')
                self.raw_compound_amount_comp = None
        elif amount := self.raw_amount_comp:
            if value:  # Amount
                amount.raw_currency = value
            else:  # Amount - Currency -> Number
                self.raw_number_comp = copy.deepcopy(amount.raw_number)
                self.raw_amount_comp = None
        else:  # Currency
            self.raw_currency_comp = value

    raw_date = raw_date_comp
    raw_label = raw_label_comp
    raw_asterisk = raw_asterisk_comp

    number_per = internal.optional_decimal_property(raw_number_per, NumberExpr)
    number_total = internal.optional_decimal_property(raw_number_total, NumberExpr)
    currency = internal.optional_string_property(raw_currency, Currency)
    date = internal.optional_date_property(raw_date, Date)
    label = internal.optional_string_property(raw_label, EscapedString)

    @property
    def merge(self) -> bool:
        return self.raw_asterisk is not None

    @merge.setter
    def merge(self, value: bool) -> None:
        current = self.merge
        if current and not value:
            self.raw_asterisk = None
        elif not current and value:
            self.raw_asterisk = Asterisk.from_default()

    @classmethod
    def from_value(
            cls: Type[_Self],
            number_per: Optional[decimal.Decimal],
            number_total: Optional[decimal.Decimal],
            currency: Optional[str],
            date: Optional[datetime.date] = None,
            label: Optional[str] = None,
            merge: bool = False,
    ) -> _Self:
        type_: Type[UnitCost | TotalCost]
        comps: list[CostComponent] = []
        if number_per is not None and number_total is not None:  # CompoundAmount
            if currency is None:
                raise ValueError('Cannot set both number_per and number_total without a currency.')
            type_ = UnitCost
            comps.append(CompoundAmount.from_children(
                NumberExpr.from_value(number_per),
                NumberExpr.from_value(number_total),
                Currency.from_value(currency)))
        elif number_per is not None:
            type_ = UnitCost
            if currency is None:
                comps.append(NumberExpr.from_value(number_per))
            else:
                comps.append(Amount.from_children(
                    NumberExpr.from_value(number_per),
                    Currency.from_value(currency)))
        elif number_total is not None:
            type_ = TotalCost
            if currency is None:
                comps.append(NumberExpr.from_value(number_total))
            else:
                comps.append(Amount.from_children(
                    NumberExpr.from_value(number_total),
                    Currency.from_value(currency)))
        elif currency is not None:
            type_ = UnitCost
            comps.append(Currency.from_value(currency))
        else:
            type_ = UnitCost
        if date is not None:
            comps.append(Date.from_value(date))
        if label is not None:
            comps.append(EscapedString.from_value(label))
        if merge:
            comps.append(Asterisk.from_default())
        return cls.from_children(type_.from_children(comps))

    def _into_unit_cost(self, total_cost: TotalCost) -> None:
        self._cost = total_cost.into_unit_cost()

    def _into_total_cost(self, unit_cost: UnitCost) -> None:
        self._cost = unit_cost.into_total_cost()



from .base_property import *
from .fields import *
from .interleaving_comments import *
from .placeholder import *
from .properties import *
from .registry import *
from .repeated import *
from .spacing_accessors import *
from .surrounding_comments import *
from .token_models import *
from .value_properties import *



from typing import Any, TypeAlias, TYPE_CHECKING
from .number import Number
if TYPE_CHECKING:
    from .number_paren_expr import NumberParenExpr
    from .number_unary_expr import NumberUnaryExpr
else:
    NumberParenExpr = Any
    NumberUnaryExpr = Any

NumberAtomExpr: TypeAlias = Number | NumberParenExpr | NumberUnaryExpr



import datetime
import decimal
import abc
import itertools
from typing import Any, Callable, Collection, Generic, Iterable, Iterator, MutableSequence, Optional, Type, TypeGuard, TypeVar, cast, overload
from .. import base
from . import indexes, base_property, properties


_V = TypeVar('_V')
_M = TypeVar('_M', bound=base.RawModel)
_M2 = TypeVar('_M2', bound=base.RawModel)
_U = TypeVar('_U', bound=base.RawTreeModel)
_SV = TypeVar('_SV', bound='RWValue[str]')
_ISV = TypeVar('_ISV', bound='RWValueWithIndent[str]')
_DV = TypeVar('_DV', bound='RWValue[decimal.Decimal]')
_DateV = TypeVar('_DateV', bound='RWValue[datetime.date]')
_SelfRWValue = TypeVar('_SelfRWValue', bound='RWValue')
_SelfRWValueWithIndent = TypeVar('_SelfRWValueWithIndent', bound='RWValueWithIndent')


class RWValue(base.RawModel, abc.ABC, Generic[_V]):
    @property
    def value(self) -> _V:
        raise NotImplementedError()

    @value.setter
    def value(self, value: _V) -> None:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def from_value(cls: Type[_SelfRWValue], value: _V) -> _SelfRWValue:
        raise NotImplementedError()


class RWValueWithIndent(RWValue[_V]):
    @property
    def indent(self) -> str:
        raise NotImplementedError()

    @indent.setter
    def indent(self, indent: str) -> None:
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def from_value(cls: Type[_SelfRWValueWithIndent], value: _V, *, indent: str = '') -> _SelfRWValueWithIndent:
        raise NotImplementedError()


class required_value_property(Generic[_V, _U]):
    def __init__(self, inner_property: base_property.base_ro_property[RWValue[_V], _U]):
        self._inner_property = inner_property

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> _V:
        return self._inner_property.__get__(instance, owner).value
    
    def __set__(self, instance: _U, value: _V) -> None:
        self._inner_property.__get__(instance).value = value


class optional_string_property(Generic[_SV]):
    def __init__(self, inner_property: base_property.base_rw_property[Optional[_SV], _U], inner_type: Type[_SV]):
        self._inner_property = inner_property
        self._inner_type = inner_type

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> Optional[str]:
        s = self._inner_property.__get__(instance, owner)
        return s.value if s is not None else None
    
    def __set__(self, instance: _U, value: Optional[str]) -> None:
        current = self._inner_property.__get__(instance)
        if current is not None and value is not None:
            current.value = value
        else:
            s = self._inner_type.from_value(value) if value is not None else None
            self._inner_property.__set__(instance, s)


class optional_indented_string_property(Generic[_ISV]):
    def __init__(
            self,
            inner_property: base_property.base_rw_property[Optional[_ISV], _U],
            inner_type: Type[_ISV],
            indent_property: base_property.base_ro_property[_SV, base.RawTreeModel]):
        self._inner_property = inner_property
        self._inner_type = inner_type
        self._indent_property = indent_property

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> Optional[str]:
        s = self._inner_property.__get__(instance, owner)
        return s.value if s is not None else None
    
    def __set__(self, instance: _U, value: Optional[str]) -> None:
        current = self._inner_property.__get__(instance)
        if current is not None and value is not None:
            current.value = value
        else:
            indent = self._indent_property.__get__(instance).value
            s = self._inner_type.from_value(value, indent=indent) if value is not None else None
            self._inner_property.__set__(instance, s)


class optional_decimal_property(Generic[_U]):
    def __init__(self, inner_property: base_property.base_rw_property[Optional[_DV], _U], inner_type: Type[_DV]):
        self._inner_property = inner_property
        self._inner_type = inner_type

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> Optional[decimal.Decimal]:
        s = self._inner_property.__get__(instance, owner)
        return s.value if s is not None else None
    
    def __set__(self, instance: _U, value: Optional[decimal.Decimal]) -> None:
        current = self._inner_property.__get__(instance)
        if current is not None and value is not None:
            current.value = value
        else:
            s = self._inner_type.from_value(value) if value is not None else None
            self._inner_property.__set__(instance, s)


class optional_date_property(Generic[_U]):
    def __init__(self, inner_property: base_property.base_rw_property[Optional[_DateV], _U], inner_type: Type[_DateV]):
        self._inner_property = inner_property
        self._inner_type = inner_type

    def __get__(self, instance: _U, owner: Optional[Type[_U]] = None) -> Optional[datetime.date]:
        s = self._inner_property.__get__(instance, owner)
        return s.value if s is not None else None
    
    def __set__(self, instance: _U, value: Optional[datetime.date]) -> None:
        current = self._inner_property.__get__(instance)
        if current is not None and value is not None:
            current.value = value
        else:
            s = self._inner_type.from_value(value) if value is not None else None
            self._inner_property.__set__(instance, s)


class RepeatedValueWrapper(MutableSequence[_V], Generic[_M, _V]):
    def __init__(
            self,
            raw_wrapper: properties.RepeatedNodeWrapper[_M | Any],
            raw_type: Type[_M] | tuple[Type[_M], ...],
            from_raw_type: Callable[[_M], _V],
            to_raw_type: Callable[[_V], _M],
            update_raw: Callable[[_M, _V], bool],
    ):
        self._raw_wrapper = raw_wrapper
        self._raw_type = raw_type
        self._from_raw_type = from_raw_type
        self._to_raw_type = to_raw_type
        self._update_raw = update_raw

    @property
    def indent(self) -> Optional[str]:
        return self._raw_wrapper.indent

    def _check_type(self, v: Any) -> TypeGuard[_M]:
        return isinstance(v, self._raw_type)

    def __len__(self) -> int:
        return sum(1 for _ in self)

    def __iter__(self) -> Iterator[_V]:
        return (
            self._from_raw_type(item) for item in self._raw_wrapper if self._check_type(item))

    def _filtered_items(self) -> list[tuple[int, _M]]:
        return [(i, item) for i, item in enumerate(self._raw_wrapper) if isinstance(item, self._raw_type)]

    @overload
    def __getitem__(self, index: int) -> _V:
        ...
    @overload
    def __getitem__(self, index: slice) -> list[_V]:
        ...
    def __getitem__(self, index: int | slice) -> _V | list[_V]:
        items = self._filtered_items()
        if isinstance(index, int):
            return self._from_raw_type(items[index][1])
        return [self._from_raw_type(item[1]) for item in items[index]]

    def __delitem__(self, index: int | slice) -> None:
        items = self._filtered_items()
        r = indexes.range_from_index(index, len(items))
        self._raw_wrapper.drop_many(items[i][0] for i in r)

    @overload
    def __setitem__(self, index: int, value: _V) -> None:
        ...
    @overload
    def __setitem__(self, index: slice, value: Iterable[_V]) -> None:
        ...
    def __setitem__(self, index: int | slice, value: _V | Iterable[_V]) -> None:
        items = self._filtered_items()
        if isinstance(index, int):
            values = [cast(_V, value)]
        else:
            assert isinstance(value, Iterable)
            values = list(value)
        r = indexes.range_from_index(index, len(items))
        items_to_update = items[indexes.slice_from_range(r)]
        # We don't allow assignment with distinct length here because the underlying models may not be consecutive.
        if len(items_to_update) != len(values):
            raise ValueError(f'attempt to assign sequence of size {len(values)} to extended slice of size {len(items_to_update)}')
        for item, value in zip(items_to_update, values):
            if not self._update_raw(item[1], value):
                self._raw_wrapper[item[0]] = self._to_raw_type(value)

    def insert(self, index: int, value: _V) -> None:
        items = self._filtered_items()
        if index >= len(items):
            underlying_index = len(self._raw_wrapper)
        elif index < -len(items):
            underlying_index = 0
        else:
            underlying_index = items[index][0]
        self._raw_wrapper.insert(underlying_index, self._to_raw_type(value))

    def append(self, value: _V) -> None:
        self._raw_wrapper.append(self._to_raw_type(value))

    def clear(self) -> None:
        items = self._filtered_items()
        self._raw_wrapper.drop_many(item[0] for item in items)

    def extend(self, values: Iterable[_V]) -> None:
        self._raw_wrapper.extend(self._to_raw_type(value) for value in values)

    def pop(self, index: int = -1) -> _V:
        items = self._filtered_items()
        if not items:
            raise IndexError('pop from empty list')
        if not -len(items) <= index < len(items):
            raise IndexError('pop index out of range')
        underlying_index = items[index][0]
        self._raw_wrapper.pop(underlying_index)
        return self._from_raw_type(items[index][1])

    def remove(self, value: _V) -> None:
        items = self._filtered_items()
        for i, item in items:
            if self._from_raw_type(item) == value:
                self._raw_wrapper.pop(i)
                return
        raise ValueError(f'{value!r} not found in list')

    def discard(self, value: _V) -> None:
        items = self._filtered_items()
        self._raw_wrapper.drop_many(item[0] for item in items if self._from_raw_type(item[1]) == value)

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Collection) and
            all(a == b for a, b in itertools.zip_longest(self, other)))


def _update_raw(raw_value: _SV, value: str) -> bool:
    raw_value.value = value
    return True


class repeated_string_property(properties.cached_custom_property[RepeatedValueWrapper[_SV, str], base.RawTreeModel]):
    def __init__(
            self,
            inner_property: base_property.base_ro_property[properties.RepeatedNodeWrapper[_SV | _M], base.RawTreeModel],
            inner_type: Type[_SV]):
        super().__init__(lambda instance: RepeatedValueWrapper[_SV, str](
            raw_wrapper=inner_property.__get__(instance),
            raw_type=inner_type,
            from_raw_type=lambda x: x.value,
            to_raw_type=inner_type.from_value,
            update_raw=_update_raw,
        ))


class RepeatedFilteredNodeWrapper(RepeatedValueWrapper[_M, _M]):
    def __init__(
            self,
            raw_wrapper: properties.RepeatedNodeWrapper[_M | _M2],
            type: Type[_M] | tuple[Type[_M], ...]):
        super().__init__(
            raw_wrapper=raw_wrapper,
            raw_type=type,
            from_raw_type=lambda x: x,
            to_raw_type=lambda x: x,
            update_raw=lambda _, __: False,
        )


class repeated_filtered_node_property(
        properties.cached_custom_property[RepeatedFilteredNodeWrapper[_M], base.RawTreeModel]):
    def __init__(
            self,
            inner_property: base_property.base_ro_property[properties.RepeatedNodeWrapper[_M | Any], base.RawTreeModel],
            type: Type[_M] | tuple[Type[_M], ...]):
        super().__init__(
            lambda instance: RepeatedFilteredNodeWrapper(inner_property.__get__(instance), type))



from . import internal


@internal.token_model
class InlineComment(internal.SingleValueRawTokenModel[str]):
    RULE = 'INLINE_COMMENT'

    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        return raw_text.removeprefix(';').lstrip(' ')

    @classmethod
    def _format_value(cls, value: str) -> str:
        return f'; {value}' if value else ';'



import decimal
from . import internal
from .generated import number_paren_expr
from .generated.number_paren_expr import LeftParen, RightParen


@internal.tree_model
class NumberParenExpr(number_paren_expr.NumberParenExpr):
    @property
    def value(self) -> decimal.Decimal:
        return self._inner_expr.value



import copy
import decimal
from typing import Any, Callable, Literal, NoReturn, Union, overload
from . import base
from . import internal
from . import number
from .number_add_expr import NumberAddExpr, AddOp
from .number_mul_expr import NumberMulExpr, MulOp
from .number_atom_expr import NumberAtomExpr
from .number_unary_expr import NumberUnaryExpr
from .number_paren_expr import NumberParenExpr
from .spacing import Whitespace
from .generated.number_paren_expr import LeftParen, RightParen
from .generated.number_unary_expr import UnaryOp
from .generated import number_expr

_AnyNumber = Union[int, decimal.Decimal, 'NumberExpr']
_OP_INFO = {
    '__iadd__': ('+=', False),
    '__isub__': ('-=', False),
    '__imul__': ('*=', False),
    '__itruediv__': ('/=', False),
    '__add__': ('+', False),
    '__sub__': ('-', False),
    '__mul__': ('*', False),
    '__truediv__': ('/', False),
    '__radd__': ('+', True),
    '__rsub__': ('-', True),
    '__rmul__': ('*', True),
    '__rtruediv__': ('/', True),
}


def _add_expr_from_value(value: decimal.Decimal) -> NumberAddExpr:
    number_token = number.Number.from_value(abs(value))
    token_store = base.TokenStore.from_tokens([number_token])
    atom_expr: NumberAtomExpr
    if value < 0:
        op = UnaryOp.from_raw_text('-')
        token_store.insert_before(number_token, [op])
        atom_expr = NumberUnaryExpr(token_store, op, number_token)
    else:
        atom_expr = number_token
    mul_expr = NumberMulExpr(token_store, (atom_expr,), ())
    return NumberAddExpr(token_store, (mul_expr,), ())



def _operand_type_check(
        op: Callable[['NumberExpr', 'NumberExpr'], 'NumberExpr'],
) -> Callable[['NumberExpr', object], 'NumberExpr']:
    def wrapped_op(self: 'NumberExpr', other: object) -> 'NumberExpr':
        if isinstance(other, int):
            other = NumberExpr.from_value(decimal.Decimal(other))
        elif isinstance(other, decimal.Decimal):
            other = NumberExpr.from_value(other)
        if isinstance(other, NumberExpr):
            return op(self, other)
        op_name, rev = _OP_INFO[op.__name__]
        lhs: Any
        rhs: Any
        if rev:
            lhs, rhs = other, self
        else:
            lhs, rhs = self, other
        raise TypeError(
            f'unsupported operand type(s) for {op_name}: '
            f'{type(lhs).__name__!r} and {type(rhs).__name__!r}.')
    return wrapped_op


def _wrap_paren(add_expr: NumberAddExpr) -> NumberParenExpr:
    left_paren = LeftParen.from_default()
    right_paren = RightParen.from_default()
    add_expr.token_store.insert_before(add_expr.first_token, [left_paren])
    add_expr.token_store.insert_after(add_expr.last_token, [right_paren])
    return NumberParenExpr(
        add_expr.token_store, left_paren, add_expr, right_paren)


def _as_mul_expr(expr: 'NumberExpr') -> NumberMulExpr:
    if not expr.raw_number_add_expr.raw_ops:
        return expr.raw_number_add_expr.raw_operands[0]
    paren_expr = _wrap_paren(expr.raw_number_add_expr)
    mul_expr = NumberMulExpr(expr.token_store, (paren_expr,), ())
    return mul_expr


def _as_atom_expr(expr: 'NumberExpr') -> NumberAtomExpr:
    if not expr.raw_number_add_expr.raw_ops:
        mul_expr = expr.raw_number_add_expr.raw_operands[0]
        if not mul_expr.raw_ops:
            return mul_expr.raw_operands[0]
    return _wrap_paren(expr.raw_number_add_expr)


def _unary(a: 'NumberExpr', op: Literal['+', '-']) -> NumberAddExpr:
    a = copy.deepcopy(a)
    atom_expr = _as_atom_expr(a)
    unary_op = UnaryOp.from_raw_text(op)
    a.token_store.insert_before(None, [unary_op])
    unary_expr = NumberUnaryExpr(a.token_store, unary_op, atom_expr)
    mul_expr = NumberMulExpr(a.token_store, (unary_expr,), ())
    return NumberAddExpr(a.token_store, (mul_expr,), ())


@internal.tree_model
class NumberExpr(number_expr.NumberExpr, internal.RWValue[decimal.Decimal]):
    @classmethod
    def from_value(cls, value: decimal.Decimal) -> 'NumberExpr':
        add_expr = _add_expr_from_value(value)
        return cls(add_expr.token_store, add_expr)

    @property
    def value(self) -> decimal.Decimal:
        return self._number_add_expr.value

    @value.setter
    def value(self, value: decimal.Decimal) -> None:
        self.raw_number_add_expr = _add_expr_from_value(value)

    def wrap_with_parenthesis(self) -> None:
        paren_expr = _wrap_paren(self.raw_number_add_expr)
        mul_expr = NumberMulExpr(self.token_store, (paren_expr,), ())
        add_expr = NumberAddExpr(self.token_store, (mul_expr,), ())
        self._number_add_expr = add_expr

    def _iaddsub(self: 'NumberExpr', other: 'NumberExpr', op: Literal['+', '-']) -> 'NumberExpr':
        mul_expr = _as_mul_expr(other)
        add_op = AddOp.from_raw_text(op)
        self.token_store.insert_after(self.last_token, [
            Whitespace.from_default(),
            add_op,
            Whitespace.from_default(),
            *mul_expr.detach(),
        ])
        mul_expr.reattach(self.token_store)
        add_expr = NumberAddExpr(
            self.token_store,
            self.raw_number_add_expr.raw_operands + (mul_expr,),
            self.raw_number_add_expr.raw_ops + (add_op,))
        self._number_add_expr = add_expr
        return self

    @overload
    def __iadd__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __iadd__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __iadd__(self, other: 'NumberExpr') -> 'NumberExpr':
        return self._iaddsub(other, '+')

    @overload
    def __add__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __add__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __add__(self, other: 'NumberExpr') -> 'NumberExpr':
        return copy.deepcopy(self).__iadd__(other)

    @overload
    def __radd__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __radd__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __radd__(self, other: 'NumberExpr') -> 'NumberExpr':
        return other + self

    @overload
    def __isub__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __isub__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __isub__(self, other: 'NumberExpr') -> 'NumberExpr':
        return self._iaddsub(other, '-')

    @overload
    def __sub__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __sub__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __sub__(self, other: 'NumberExpr') -> 'NumberExpr':
        return copy.deepcopy(self).__isub__(other)

    @overload
    def __rsub__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __rsub__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __rsub__(self, other: 'NumberExpr') -> 'NumberExpr':
        return other - self

    def _imuldiv(self: 'NumberExpr', other: 'NumberExpr', op: Literal['*', '/']) -> 'NumberExpr':
        self_mul_expr = _as_mul_expr(self)
        atom_expr = _as_atom_expr(other)
        mul_op = MulOp.from_raw_text(op)
        self.token_store.insert_after(self_mul_expr.last_token, [
            Whitespace.from_default(),
            mul_op,
            Whitespace.from_default(),
            *atom_expr.detach(),
        ])
        atom_expr.reattach(self.token_store)
        mul_expr = NumberMulExpr(
            self.token_store,
            self_mul_expr.raw_operands + (atom_expr,),
            self_mul_expr.raw_ops + (mul_op,))
        add_expr = NumberAddExpr(self.token_store, (mul_expr,), ())
        self._number_add_expr = add_expr
        return self

    @overload
    def __imul__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __imul__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __imul__(self, other: 'NumberExpr') -> 'NumberExpr':
        return self._imuldiv(other, '*')

    @overload
    def __mul__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __mul__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __mul__(self, other: 'NumberExpr') -> 'NumberExpr':
        return copy.deepcopy(self).__imul__(other)

    @overload
    def __rmul__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __rmul__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __rmul__(self, other: 'NumberExpr') -> 'NumberExpr':
        return other * self

    @overload
    def __itruediv__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __itruediv__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __itruediv__(self, other: 'NumberExpr') -> 'NumberExpr':
        return self._imuldiv(other, '/')

    @overload
    def __truediv__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __truediv__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __truediv__(self, other: 'NumberExpr') -> 'NumberExpr':
        return copy.deepcopy(self).__itruediv__(other)

    @overload
    def __rtruediv__(self, other: _AnyNumber) -> 'NumberExpr':  # type: ignore[misc]
        ...
    @overload
    def __rtruediv__(self, other: object) -> NoReturn:
        ...
    @_operand_type_check
    def __rtruediv__(self, other: 'NumberExpr') -> 'NumberExpr':
        return other / self

    def __pos__(self) -> 'NumberExpr':
        add_expr = _unary(self, '+')
        return type(self)(add_expr.token_store, add_expr)

    def __neg__(self) -> 'NumberExpr':
        add_expr = _unary(self, '-')
        return type(self)(add_expr.token_store, add_expr)




import abc
from typing import Type, TypeVar, final
from .. import base
from .value_properties import RWValue

_V = TypeVar('_V')
_SelfSimpleRawTokenModel = TypeVar('_SelfSimpleRawTokenModel', bound='SimpleRawTokenModel')
_SelfSingleValueRawTokenModel = TypeVar('_SelfSingleValueRawTokenModel', bound='SingleValueRawTokenModel')
_SelfDefaultRawTokenModel = TypeVar('_SelfDefaultRawTokenModel', bound='DefaultRawTokenModel')


class SimpleRawTokenModel(base.RawTokenModel):
    @final
    def __init__(self, raw_text: str) -> None:
        super().__init__(raw_text)

    def _clone(self: _SelfSimpleRawTokenModel) -> _SelfSimpleRawTokenModel:
        return type(self)(self.raw_text)


class SingleValueRawTokenModel(base.RawTokenModel, RWValue[_V]):
    @final
    def __init__(self, raw_text: str, value: _V) -> None:
        super().__init__(raw_text)
        self._value = value

    @classmethod
    def from_raw_text(cls: Type[_SelfSingleValueRawTokenModel], raw_text: str) -> _SelfSingleValueRawTokenModel:
        return cls(raw_text, cls._parse_value(raw_text))

    @classmethod
    def from_value(cls: Type[_SelfSingleValueRawTokenModel], value: _V) -> _SelfSingleValueRawTokenModel:
        return cls(cls._format_value(value), value)

    @property
    def raw_text(self) -> str:
        return super().raw_text

    @raw_text.setter
    def raw_text(self, raw_text: str) -> None:
        self._update_raw_text(raw_text)
        self._value = self._parse_value(raw_text)

    @property
    def value(self) -> _V:
        return self._value

    @value.setter
    def value(self, value: _V) -> None:
        self._value = value
        self._raw_text = self._format_value(value)

    @classmethod
    @abc.abstractmethod
    def _parse_value(cls, raw_text: str) -> _V:
        pass

    @classmethod
    @abc.abstractmethod
    def _format_value(cls, value: _V) -> str:
        pass

    def _clone(self: _SelfSingleValueRawTokenModel) -> _SelfSingleValueRawTokenModel:
        return type(self)(self.raw_text, self.value)


class SimpleSingleValueRawTokenModel(SingleValueRawTokenModel[str]):
    @classmethod
    def _parse_value(cls, raw_text: str) -> str:
        return raw_text

    @classmethod
    def _format_value(cls, value: str) -> str:
        return value


class DefaultRawTokenModel(base.RawTokenModel):
    # not using @classmethod here because it suppresses abstractmethod errors.
    @property
    @abc.abstractmethod
    def DEFAULT(self) -> str:
        ...

    @classmethod
    def from_default(cls: Type[_SelfDefaultRawTokenModel]) -> _SelfDefaultRawTokenModel:
        return cls.from_raw_text(cls.DEFAULT)  # type: ignore[arg-type]


class SimpleDefaultRawTokenModel(SimpleRawTokenModel, DefaultRawTokenModel):
    pass



import itertools
from typing import Callable, Generic, Iterable, Iterator, MutableSet, Optional, TypeVar
from .. import base
from ..spacing import Newline, Whitespace
from ..block_comment import BlockComment
from . import base_property, fields, properties, repeated
from .placeholder import Placeholder

_M = TypeVar('_M', bound=base.RawModel)
_U = TypeVar('_U', bound=base.RawTreeModel)


class _Universe(MutableSet[int]):
    def __contains__(self, item: object) -> bool:
        return True

    def __iter__(self) -> Iterator[int]:
        raise NotImplementedError()

    def __len__(self) -> int:
        return 0

    def add(self, item: int) -> None:
        pass

    def discard(self, item: int) -> None:
        pass


def _shift_ignored(
        token_store: base.TokenStore,
        first: base.RawTokenModel,
        last: base.RawTokenModel,
        *,
        backwards: bool,
) -> None:
    ignored: list[base.RawTokenModel] = []
    others: list[base.RawTokenModel] = []
    for token in token_store.iter(first, last):
        if isinstance(token, Placeholder):
            ignored.append(token)
        else:
            others.append(token)
    if not ignored:
        return None
    if backwards:
        token_store.splice(ignored + others ,first, last)
    else:
        token_store.splice(others + ignored, first, last)


class _CommentClaimer(Generic[_M]):
    def __init__(
            self,
            repeated: repeated.Repeated[_M | BlockComment],
            model: base.RawTreeModel,
            comments: Optional[Iterable[BlockComment]]):
        self._repeated = repeated
        self._model = model
        self._comments_to_claim = (
            {id(comment) for comment in comments} if comments is not None else _Universe())

    def _find_inner(self) -> Iterator[_M | BlockComment]:
        token_store = self._repeated.token_store
        # inclusive
        starts = itertools.chain(
            [self._repeated.first_token],
            map(lambda x: token_store.get_next(x.last_token), self._repeated.items))
        # exclusive
        ends = map(lambda x: x.first_token, self._repeated.items)
        for start, end, item in zip(starts, ends, self._repeated.items):
            token = start
            while token is not None and token is not end:
                prev_token, token = token, self._repeated.token_store.get_next(token)
                if not isinstance(prev_token, BlockComment):
                    continue
                if id(prev_token) not in self._comments_to_claim:
                    continue
                if prev_token.claimed:
                    continue
                self._comments_to_claim.discard(id(prev_token))
                yield prev_token
            yield item
            self._comments_to_claim.discard(id(item))
    
    def _find_outer(
            self,
            start: base.RawTokenModel,
            succ: Callable[[base.RawTokenModel], Optional[base.RawTokenModel]],
            *,
            limit: base.RawTokenModel,
    ) -> Iterator[BlockComment]:
        prev, token = start, succ(start)
        while prev is not limit and token is not None:
            if isinstance(token, Newline | Whitespace) or not token.raw_text:
                pass
            elif isinstance(token, BlockComment):
                if token.claimed:
                    break
                if id(token) in self._comments_to_claim:
                    yield token
            else:
                break
            prev, token = token, succ(token)

    def claim(self) -> list[BlockComment]:
        comments_before = list(self._find_outer(
            self._repeated.first_token,
            self._repeated.token_store.get_prev,
            limit=self._model.first_token))
        comments_before.reverse()
        items_inner = list(self._find_inner())
        comments_after = list(self._find_outer(
            self._repeated.last_token,
            self._repeated.token_store.get_next,
            limit=self._model.last_token))

        if self._comments_to_claim:
            raise ValueError(f'{len(self._comments_to_claim)} comment(s) not found.')

        if comments_before:
            first = self._repeated.token_store.get_prev(self._repeated.first_token)
            assert first is not None
            _shift_ignored(
                self._repeated.token_store, first, comments_before[0], backwards=True)

        if comments_after:
            first = self._repeated.token_store.get_next(self._repeated.last_token)
            assert first is not None
            _shift_ignored(
                self._repeated.token_store, first, comments_after[-1], backwards=False)

        items = list(itertools.chain(comments_before, items_inner, comments_after))
        comments = []
        for item in items:
            if isinstance(item, BlockComment):
                comments.append(item)
                item.claimed = True
        self._repeated.items[:] = items
        return comments


class RepeatedNodeWithInterleavingCommentsWrapper(properties.RepeatedNodeWrapper[_M | BlockComment]):

    def __init__(
            self,
            repeated: repeated.Repeated[_M | BlockComment],
            field: fields.repeated_field,
            model: base.RawTreeModel,
    ) -> None:
        super().__init__(repeated, field)
        self._model = model

    def claim_interleaving_comments(
            self,
            comments: Optional[Iterable[BlockComment]] = None,
    ) -> tuple[BlockComment, ...]:
        return tuple(_CommentClaimer(self._repeated, self._model, comments).claim())

    def unclaim_interleaving_comments(
            self,
            comments: Optional[Iterable[BlockComment]] = None,
    ) -> tuple[BlockComment, ...]:
        unclaimed_comments = []
        comment_set = (
            {id(comment) for comment in comments} if comments is not None else None)
        items: list[_M | BlockComment] = []
        for item in self._repeated.items:
            if not isinstance(item, BlockComment):
                items.append(item)
                continue
            if comment_set is not None and id(item) not in comment_set:
                items.append(item)
                continue
            if comment_set is not None:
                comment_set.discard(id(item))
            item.claimed = False
            unclaimed_comments.append(item)
        if comment_set:
            raise ValueError(f'{len(comment_set)} comment(s) not found.')
        self._repeated.items[:] = items
        return tuple(unclaimed_comments)

    def auto_claim_comments(self) -> None:
        super().auto_claim_comments()
        self.claim_interleaving_comments()


class repeated_node_with_interleaving_comments_property(
        base_property.base_ro_property[RepeatedNodeWithInterleavingCommentsWrapper[_M], base.RawTreeModel]):
    def __init__(self, inner_field: fields.repeated_field[_M | BlockComment]) -> None:
        super().__init__()
        self._inner_field = inner_field

    def __set_name__(self, owner: base.RawTreeModel, name: str) -> None:
        self._attr = name

    def _get(self, instance: _U) -> RepeatedNodeWithInterleavingCommentsWrapper[_M]:
        wrapper = instance.__dict__.get(self._attr)
        if wrapper is None:
            repeated = self._inner_field.__get__(instance)
            wrapper = RepeatedNodeWithInterleavingCommentsWrapper(repeated, self._inner_field, instance)
            instance.__dict__[self._attr] = wrapper
        return wrapper

    def __set__(self, instance: _U, value: RepeatedNodeWithInterleavingCommentsWrapper[_M]) -> None:
        repeated = self._inner_field.__get__(instance)
        properties.replace_node(repeated, value.repeated)
        self._inner_field.__set__(instance, value.repeated)
        instance.__dict__[self._attr] = value



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import decimal
from typing import Optional, Type, TypeVar, final
from .. import base, internal
from ..currency import Currency
from ..number_expr import NumberExpr
from ..spacing import Whitespace

_Self = TypeVar('_Self', bound='CompoundAmount')


@internal.token_model
class Hash(internal.SimpleDefaultRawTokenModel):
    RULE = 'HASH'
    DEFAULT = '#'


@internal.tree_model
class CompoundAmount(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'compound_amount'

    _number_per = internal.optional_right_field[NumberExpr](separators=(Whitespace.from_default(),))
    _hash = internal.required_field[Hash]()
    _number_total = internal.optional_left_field[NumberExpr](separators=(Whitespace.from_default(),))
    _currency = internal.required_field[Currency]()

    @internal.custom_property
    def _number_per_pivot(self) -> base.RawTokenModel:
        return self._hash.first_token

    @internal.custom_property
    def _number_total_pivot(self) -> base.RawTokenModel:
        return self._hash.last_token

    raw_number_per = internal.optional_node_property(_number_per, _number_per_pivot)
    raw_number_total = internal.optional_node_property(_number_total, _number_total_pivot)
    raw_currency = internal.required_node_property(_currency)

    number_per = internal.optional_decimal_property(raw_number_per, NumberExpr)
    number_total = internal.optional_decimal_property(raw_number_total, NumberExpr)
    currency = internal.required_value_property(raw_currency)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            number_per: Optional[NumberExpr],
            hash: Hash,
            number_total: Optional[NumberExpr],
            currency: Currency,
    ):
        super().__init__(token_store)
        self._number_per = number_per
        self._hash = hash
        self._number_total = number_total
        self._currency = currency

    @property
    def first_token(self) -> base.RawTokenModel:
        return (self._number_per and self._number_per.first_token) or self._hash.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._currency.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._number_per.clone(self._number_per, token_store, token_transformer),
            type(self)._hash.clone(self._hash, token_store, token_transformer),
            type(self)._number_total.clone(self._number_total, token_store, token_transformer),
            type(self)._currency.clone(self._currency, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._number_per = type(self)._number_per.reattach(self._number_per, token_store, token_transformer)
        self._hash = type(self)._hash.reattach(self._hash, token_store, token_transformer)
        self._number_total = type(self)._number_total.reattach(self._number_total, token_store, token_transformer)
        self._currency = type(self)._currency.reattach(self._currency, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, CompoundAmount)
            and self._number_per == other._number_per
            and self._hash == other._hash
            and self._number_total == other._number_total
            and self._currency == other._currency
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            number_per: Optional[NumberExpr],
            number_total: Optional[NumberExpr],
            currency: Currency,
    ) -> _Self:
        hash = Hash.from_default()
        tokens = [
            *cls._number_per.detach_with_separators(number_per),
            *hash.detach(),
            *cls._number_total.detach_with_separators(number_total),
            Whitespace.from_default(),
            *currency.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._number_per.reattach(number_per, token_store)
        cls._hash.reattach(hash, token_store)
        cls._number_total.reattach(number_total, token_store)
        cls._currency.reattach(currency, token_store)
        return cls(token_store, number_per, hash, number_total, currency)

    @classmethod
    def from_value(
            cls: Type[_Self],
            number_per: Optional[decimal.Decimal],
            number_total: Optional[decimal.Decimal],
            currency: str,
    ) -> _Self:
        return cls.from_children(
            number_per=NumberExpr.from_value(number_per) if number_per is not None else None,
            number_total=NumberExpr.from_value(number_total) if number_total is not None else None,
            currency=Currency.from_value(currency),
        )

    def auto_claim_comments(self) -> None:
        type(self)._currency.auto_claim_comments(self._currency)
        type(self)._number_total.auto_claim_comments(self._number_total)
        type(self)._number_per.auto_claim_comments(self._number_per)



from .internal import registry as _registry
from .internal import base_token_models as _base_token_models


@_registry.token_model
class Newline(_base_token_models.SimpleDefaultRawTokenModel):
    RULE = '_NEWLINE'
    DEFAULT = '\n'


@_registry.token_model
class Whitespace(_base_token_models.SimpleDefaultRawTokenModel):
    RULE = 'WHITESPACE'
    DEFAULT = ' '



import decimal
from typing import Type, TypeVar
from . import internal
from .generated import tolerance
from .generated.tolerance import Tilde

_Self = TypeVar('_Self', bound='Tolerance')


@internal.tree_model
class Tolerance(tolerance.Tolerance, internal.RWValue[decimal.Decimal]):

    @property
    def value(self) -> decimal.Decimal:
        return self.raw_number.value

    @value.setter
    def value(self, value: decimal.Decimal) -> None:
        self.raw_number.value = value

    @classmethod
    def from_value(
            cls: Type[_Self],
            number: decimal.Decimal,
    ) -> _Self:
        return super().from_value(number)



from . import internal


@internal.token_model
class Bool(internal.SingleValueRawTokenModel[bool]):
    RULE = 'BOOL'

    @classmethod
    def _parse_value(cls, raw_text: str) -> bool:
        return {
            'TRUE': True,
            'FALSE': False,
        }[raw_text]

    @classmethod
    def _format_value(cls, value: bool) -> str:
        return str(value).upper()



from typing import Callable, Optional
from . import fields
from .placeholder import Placeholder
from .. import base
from ..spacing import Newline
from ..block_comment import BlockComment


def _take_ignored(
        token: Optional[base.RawTokenModel],
        succ: Callable[[base.RawTokenModel], Optional[base.RawTokenModel]],
        ignored: list[base.RawTokenModel],
) -> Optional[base.RawTokenModel]:
    while isinstance(token, Placeholder):
        ignored.append(token)
        token = succ(token)
    return token


def _claim_comment(
    current: Optional[BlockComment],
    token_store: base.TokenStore,
    start: base.RawTokenModel,
    *,
    backwards: bool,
    ignore_if_already_claimed: bool,
) -> Optional[BlockComment]:
    if current is not None:
        return current

    ignored: list[base.RawTokenModel] = []
    if backwards:
        succ = token_store.get_prev
    else:
        succ = token_store.get_next

    first = succ(start)
    if first is None:
        return None

    newline = _take_ignored(first, succ, ignored)
    if not isinstance(newline, Newline):
        return None
    comment = _take_ignored(succ(newline), succ, ignored)
    if not isinstance(comment, BlockComment):
        return None

    if comment.claimed:
        if ignore_if_already_claimed:
            return None
        raise ValueError('Comment already claimed.')
    comment.claimed = True
    if ignored:
        if backwards:
            token_store.splice(
                [*reversed(ignored), comment, newline], comment, first)
        else:
            token_store.splice(
                [newline, comment, *ignored], first, comment)
    return comment


class SurroundingCommentsMixin(base.RawTreeModel):
    _leading_comment = fields.optional_right_field[BlockComment](separators=(Newline.from_default(),))
    _trailing_comment = fields.optional_left_field[BlockComment](separators=(Newline.from_default(),))

    def claim_leading_comment(self, *, ignore_if_already_claimed: bool = False) -> Optional[BlockComment]:
        self._leading_comment = _claim_comment(
            self._leading_comment,
            self.token_store,
            self.first_token,
            backwards=True,
            ignore_if_already_claimed=ignore_if_already_claimed)
        return self._leading_comment

    def unclaim_leading_comment(self) -> Optional[BlockComment]:
        current = self._leading_comment
        if current is not None:
            current.claimed = False
            self._leading_comment = None
        return current

    def claim_trailing_comment(self, *, ignore_if_already_claimed: bool = False) -> Optional[BlockComment]:
        self._trailing_comment = _claim_comment(
            self._trailing_comment,
            self.token_store,
            self.last_token,
            backwards=False,
            ignore_if_already_claimed=ignore_if_already_claimed)
        return self._trailing_comment

    def unclaim_trailing_comment(self) -> Optional[BlockComment]:
        current = self._trailing_comment
        if current is not None:
            current.claimed = False
            self._trailing_comment = None
        return current



# DO NOT EDIT
# This file is automatically generated by autobean_refactor.modelgen.

import decimal
from typing import Type, TypeVar, final
from .. import base, internal
from ..number_expr import NumberExpr
from ..spacing import Whitespace

_Self = TypeVar('_Self', bound='Tolerance')


@internal.token_model
class Tilde(internal.SimpleDefaultRawTokenModel):
    RULE = 'TILDE'
    DEFAULT = '~'


@internal.tree_model
class Tolerance(base.RawTreeModel, internal.SpacingAccessorsMixin):
    RULE = 'tolerance'

    _tilde = internal.required_field[Tilde]()
    _number = internal.required_field[NumberExpr]()

    raw_number = internal.required_node_property(_number)

    number = internal.required_value_property(raw_number)

    @final
    def __init__(
            self,
            token_store: base.TokenStore,
            tilde: Tilde,
            number: NumberExpr,
    ):
        super().__init__(token_store)
        self._tilde = tilde
        self._number = number

    @property
    def first_token(self) -> base.RawTokenModel:
        return self._tilde.first_token

    @property
    def last_token(self) -> base.RawTokenModel:
        return self._number.last_token

    def clone(self: _Self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> _Self:
        return type(self)(
            token_store,
            type(self)._tilde.clone(self._tilde, token_store, token_transformer),
            type(self)._number.clone(self._number, token_store, token_transformer),
        )

    def _reattach(self, token_store: base.TokenStore, token_transformer: base.TokenTransformer) -> None:
        self._token_store = token_store
        self._tilde = type(self)._tilde.reattach(self._tilde, token_store, token_transformer)
        self._number = type(self)._number.reattach(self._number, token_store, token_transformer)

    def _eq(self, other: base.RawTreeModel) -> bool:
        return (
            isinstance(other, Tolerance)
            and self._tilde == other._tilde
            and self._number == other._number
        )

    @classmethod
    def from_children(
            cls: Type[_Self],
            number: NumberExpr,
    ) -> _Self:
        tilde = Tilde.from_default()
        tokens = [
            *tilde.detach(),
            Whitespace.from_default(),
            *number.detach(),
        ]
        token_store = base.TokenStore.from_tokens(tokens)
        cls._tilde.reattach(tilde, token_store)
        cls._number.reattach(number, token_store)
        return cls(token_store, tilde, number)

    @classmethod
    def from_value(
            cls: Type[_Self],
            number: decimal.Decimal,
    ) -> _Self:
        return cls.from_children(
            number=NumberExpr.from_value(number),
        )

    def auto_claim_comments(self) -> None:
        type(self)._number.auto_claim_comments(self._number)



import re
import datetime
from . import internal


@internal.token_model
class Date(internal.SingleValueRawTokenModel[datetime.date]):
    RULE = 'DATE'

    @classmethod
    def _parse_value(cls, raw_text: str) -> datetime.date:
        y, m, d = map(int, re.split('[-/]', raw_text))
        return datetime.date(y, m, d)

    @classmethod
    def _format_value(cls, value: datetime.date) -> str:
        return value.strftime('%Y-%m-%d')



from . import internal


@internal.token_model
class Eol(internal.SimpleDefaultRawTokenModel):
    RULE = 'EOL'
    DEFAULT = ''


@internal.token_model
class Indent(internal.SimpleSingleValueRawTokenModel, internal.DefaultRawTokenModel):
    RULE = 'INDENT'
    DEFAULT = ' ' * 4


@internal.token_model
class DedentMark(internal.SimpleDefaultRawTokenModel):
    RULE = 'DEDENT_MARK'
    DEFAULT = ''


@internal.token_model
class Comma(internal.SimpleDefaultRawTokenModel):
    RULE = '_COMMA'
    DEFAULT = ','


@internal.token_model
class Asterisk(internal.SimpleDefaultRawTokenModel):
    RULE = 'ASTERISK'
    DEFAULT = '*'
